structure Inference = 
struct
  structure L = LambdaPlusPlus
  exception TypeError of string

  (*  occurs : int -> L.typ -> bool
   *  occurs a t ==> whether the type variable a occurs in the type t       *)
  fun occurs a = (fn
    L.TypeVariable b => b = a
  | (L.Arrow (t1, t2) | L.Plus (t1, t2) | L.Times (t1, t2)) =>
      occurs a t1 orelse occurs a t2
  )

  (*  subst : (int * L.typ) * L.typ -> L.typ
   *  subst ((a, t), tt) = substitutes all instances of the type variable a
   *    in tt with the type t
   *  not curried so it can be used with foldl                              *)
  fun subst (a, t) tt =
    case tt of
      L.TypeVariable b => if b = a then t else tt
    | L.Arrow (t1, t2) => L.Arrow (subst (a, t) t1, subst (a, t) t2)
    | L.Plus (t1, t2) => L.Plus (subst (a, t) t1, subst (a, t) t2)
    | L.Times (t1, t2) => L.Times (subst (a, t) t1, subst (a, t) t2)
  fun subst' (at, tt) = subst at tt (* uncurried version *)

  (*  unify : L.typ * L.typ -> (int * L.typ) list
   *  unify (t1, t2) raises TypeError if t1 and t2 can't be unified
   *  unify (t1, t2) ==> substitutions if t1 and t2 can be unified,
   *    where substitutions is a most general unifier, and for all elements
   *    (a, t) in substitutions, t does not contain any type variable which
   *    was bound by an earlier element of substitutions.                   *)
  val rec unify = (fn
    ((L.TypeVariable a, t) | (t, L.TypeVariable a)) =>
      if occurs a t andalso t <> L.TypeVariable a
      then raise TypeError "circularity"
      else [(a, t)]
  | ((L.Arrow (t1, t2), L.Arrow (t1', t2'))
    |(L.Plus (t1, t2), L.Plus (t1', t2'))
    |(L.Times (t1, t2), L.Times (t1', t2'))) =>
      let
        val s1 = unify (t1, t1')
        val s2 = unify (foldl subst' t2 s1, foldl subst' t2' s1)
      in
        s1 @ s2
      end
  | _ => raise TypeError "head symbol conflict"
  )

  (*  TC = type context. Will be used to store types of type-level variables
   *    that have been "unified away".                                      
   *  EC = expression context. Will be used to store types of expression-level
   *    variables.                                                          *)
  structure TC = IntRedBlackDict
  structure EC = StringRedBlackDict
  (*  ctx = a map tctx from type variables to types and
   *        a map ectx from expression variables to types.
   *  invariant: we will say that any type variable which tctx maps to
   *    something has been "unified away". Any type variable which has been
   *    unified away must not occur in any type mapped to by tctx.
   *)
  type tctx = L.typ TC.dict
  type ectx = L.typ EC.dict

  (*  apply_tctx : tctx -> L.typ -> L.typ
      apply_tctx tctx t ==> t, but all occurrences of type variables in t
        which tctx has a mapping for are substituted by those mappings.     *)
  fun apply_tctx tctx = (fn
    L.TypeVariable a => (
      case TC.find tctx a of
        NONE => L.TypeVariable a
      | SOME t => t
    )
  | L.Arrow (t1, t2) => L.Arrow (apply_tctx tctx t1, apply_tctx tctx t2)
  | L.Plus (t1, t2) => L.Plus (apply_tctx tctx t1, apply_tctx tctx t2)
  | L.Times (t1, t2) => L.Times (apply_tctx tctx t1, apply_tctx tctx t2)
  )
  
  (*  unify_tctx : L.typ * L.typ -> tctx -> tctx
   *  unify_tctx (t1, t2) tctx raises TypeError if t1 and t2 can not be
   *    unified.
   *  unify_tctx (t1, t2) tctx ==> tctx' such that
   *    apply_tctx tctx' t1 = apply_tctx tctx' t2, and
   *    tctx' is well-formed context.
   *  precondition: t1 and t2 do not contain any variables that have already
   *    been unified away.
   *)
  local
    fun subst_tctx ((a, t), tctx) =
      TC.insert (TC.map (subst (a, t)) tctx) a t
  in
    fun unify_tctx t1t2 tctx = foldl subst_tctx tctx (unify t1t2)
  end

  (*  fresh : unit -> int
   *  fresh () ==> a unique int every time it's called.
   *
   *  fresh_t : unit -> L.typ
   *  fresh_t () ==> a unique type variable every time it's called.         *)
  local
    val n = ref 0
  in
    fun fresh () = (
      n := !n + 1;
      !n
    )

    fun fresh_t () = L.TypeVariable (fresh ())
  end
  
  (*  infer_help : tctx -> ectx -> L.exp -> L.typ * tctx
   *  infer_help tctx ectx e raises TypeError if e can not be well-typed
   *  infer_help tctx ectx e ==> (t, tctx') where t is the most general type e can
   *    have, and tctx' is a type context updated with all unifications that
   *    took place while inferring the type.
   *  precondition: tctx is well-formed
   *  postcondition: t must not contain any type variables that are unified
   *    away in tctx'.
   *)
  fun infer_help tctx ectx = (fn
    L.Variable x => (
      case EC.find ectx x of
        NONE => raise TypeError ("unbound variable: " ^ x)
      | SOME t => (apply_tctx tctx t, tctx)
    )
  | L.Lambda (x, e) =>
    let
      val t1 = fresh_t ()
      val (t2, tctx') = infer_help tctx (EC.insert ectx x t1) e
    in
      (L.Arrow (apply_tctx tctx' t1, t2), tctx')
    end
  | L.Apply (e1, e2) =>
    let
      val (t1, t2) = (fresh_t (), fresh_t ())
      val (t_e1, tctx1) = infer_help tctx ectx e1
      val (t_e2, tctx2) = infer_help tctx1 ectx e2
      val tctx3 = unify_tctx (L.Arrow (t1, t2), t_e1) tctx2
      val tctx4 = unify_tctx (apply_tctx tctx3 t1, t_e2) tctx3
    in
      (apply_tctx tctx4 t2, tctx4)
    end
  | L.Tuple (e1, e2) =>
    let
      val (t_e1, tctx') = infer_help tctx ectx e1
      val (t_e2, tctx'') = infer_help tctx' ectx e2
    in
      (L.Times (apply_tctx tctx'' t_e1, t_e2), tctx'')
    end
  | L.First e =>
    let
      val (t1, t2) = (fresh_t (), fresh_t ())
      val (t, tctx') = infer_help tctx ectx e
      val tctx'' = unify_tctx (L.Times (t1, t2), t) tctx'
    in
      (apply_tctx tctx'' t1, tctx'')
    end
  | L.Second e =>
    let
      val (t1, t2) = (fresh_t (), fresh_t ())
      val (t, tctx') = infer_help tctx ectx e
      val tctx'' = unify_tctx (L.Times (t1, t2), t) tctx'
    in
      (apply_tctx tctx'' t2, tctx'')
    end
  | L.Left e =>
    let
      val t2 = fresh_t ()
      val (t1, tctx') = infer_help tctx ectx e
    in
      (L.Plus (t1, t2), tctx')
    end
  | L.Right e =>
    let
      val t1 = fresh_t ()
      val (t2, ctx') = infer_help tctx ectx e
    in
      (L.Plus (t1, t2), ctx')
    end
  | L.Case (e, (x1, e1), (x2, e2)) =>
    let
      val (t, tctx1) = infer_help tctx ectx e
      val (t1, t2) = (fresh_t (), fresh_t ())
      val tctx2 = unify_tctx (L.Plus (t1, t2), t) tctx1
      val (t1_res, tctx3) = infer_help tctx2 (EC.insert ectx x1 t1) e1
      val (t2_res, tctx4) = infer_help tctx3 (EC.insert ectx x2 t2) e2
      val tctx5 = unify_tctx (t1_res, t2_res) tctx4
    in
      (apply_tctx tctx5 t1_res, tctx5)
    end
  )
  
  val infer = #1 o infer_help TC.empty EC.empty

end

structure ReverseInference = 
struct
  structure L = LambdaPlusPlus

  infix 4 >>=
  fun NONE >>= f = NONE
    | (SOME x) >>= f = f x
  val return = SOME
  
  local
    val n = ref ~1
    
    fun ith_name i =
      if i < 26
      then Char.toString (chr (i + 97))
      else ith_name (i div 26) ^ ith_name (i mod 26)
  in
    fun fresh_string () = (
      n := !n + 1;
      ith_name (!n)
    )
  end

  (* all ways to pick one element out of a list *)
  fun plucks [] = []
    | plucks (x::xs) = (x, xs) :: map (fn (y, ys) => (y, x::ys)) (plucks xs)

  fun right neg ctx = (fn
    L.Times (t1, t2) =>
      right neg ctx t1 >>= (fn e1 =>
        right neg ctx t2 >>= (fn e2 =>
          return (L.Tuple (e1, e2))
        )
      )
  | L.Arrow (t1, t2) =>
    let val x = fresh_string () in
      right neg ((t1, L.Variable x)::ctx) t2 >>= (fn e =>
        return (L.Lambda (x, e))
      )
    end
  | (t as (L.TypeVariable _ | L.Plus _)) => left neg ctx t
  )

  and left neg ((c, e)::ctx) t = (
      case c of
        L.Times (t1, t2) =>
          left neg ((t1, L.First e)::(t2, L.Second e)::ctx) t
      | L.Plus (t1, t2) =>
        let val (x1, x2) = (fresh_string (), fresh_string ()) in
          left neg ((t1, L.Variable x1)::ctx) t >>= (fn e1 =>
            left neg ((t2, L.Variable x2)::ctx) t >>= (fn e2 =>
              return (L.Case (e, (x1, e1), (x2, e2)))
            )
          )
        end
      | (L.TypeVariable _ | L.Arrow _) => left ((c, e)::neg) ctx t
    )
    | left neg [] t =
      let
        val left_arrows =
          List.mapPartial
            (fn ((L.Arrow (t1, t2), e), neg') => SOME ((t1, t2), e, neg')
              | _ => NONE)
            (plucks neg)
        val left_arrow_attempts =
          List.mapPartial (fn
            ((L.Plus (a1, a2), t2), e, neg') =>
              let
                val (x1, x2) = (fresh_string (), fresh_string ())
                val e1 = L.Lambda (x1, L.Apply (e, L.Left (L.Variable x1)))
                val e2 = L.Lambda (x2, L.Apply (e, L.Right (L.Variable x2)))
              in
                left ((L.Arrow (a1, t2), e1)::(L.Arrow (a2, t2), e2)::neg') [] t
              end
          | ((L.Times (a1, a2), t2), e, neg') =>
              let
                val (x1, x2) = (fresh_string (), fresh_string ())
                val e' = L.Lambda (x1,
                  L.Lambda (x2,
                    L.Apply (e, L.Tuple (L.Variable x1, L.Variable x2))
                  )
                )
              in
                left ((L.Arrow (a1, L.Arrow (a2, t2)), e')::neg') [] t
              end
          | ((L.TypeVariable a, t2), e, neg') =>
              List.find
                (fn (L.TypeVariable b, _) => a = b | _ => false)
                neg' >>= (fn (_, e') =>
                  left neg' [(t2, L.Apply (e, e'))] t
              )
          | ((L.Arrow (a1, a2), t2), e, neg') =>
              let
                val (x1, x1') = (fresh_string (), fresh_string ())
                val e1 = L.Lambda (x1,
                  L.Apply (e, L.Lambda (x1', L.Variable x1))
                )
              in
                right ((L.Arrow (a2, t2), e1)::neg') [] (L.Arrow (a1, a2))
                >>= (fn e2 => left neg' [(t2, L.Apply (e, e2))] t)
              end
          ) left_arrows
      in
        case left_arrow_attempts of
          e::_ => SOME e
        | [] =>
          case t of
            L.Plus (t1, t2) => (
              case left neg [] t1 of
                SOME e => SOME (L.Left e)
              | NONE => (
                case left neg [] t2 of
                  SOME e => SOME (L.Right e)
                | NONE => NONE
              )
            )
          | L.TypeVariable a =>
              List.find
                (fn (L.TypeVariable b, _) => a = b | _ => false)
                neg >>= (SOME o #2)
          | _ => NONE
      end


  val reverse_infer = right [] []
end